//-----------------------------------------------------------------------------------
// File : asdxModelUtil.h
// Desc : Asura Model Utility for DirectX 11
// Copyright(c) Project Asura. All right reserved.
//-----------------------------------------------------------------------------------

//-----------------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------------
#include "asDef.h"
#include "asModel.h"
#include "asTimer.h"
#include <d3d11.h>
#include <d3dx10math.h>

namespace Asura {
namespace Util {

//////////////////////////////////////////////////////////////////////////
// ModelUtil class
//////////////////////////////////////////////////////////////////////////
class ModelUtil
{
	//--------------------------------------------------------------------
	// list of friend classes and methods
	//--------------------------------------------------------------------
	/* NOTHING */

private:
	//--------------------------------------------------------------------
	// private variables
	//--------------------------------------------------------------------
	/* NOTHING */

	//--------------------------------------------------------------------
	// private methods
	//--------------------------------------------------------------------
	/* NOTHING */

protected:
	//--------------------------------------------------------------------
	// protected variables
	//--------------------------------------------------------------------
	/* NOTHING */

	//--------------------------------------------------------------------
	// protected methods
	//--------------------------------------------------------------------
	/* NOTHING */

public:
	//--------------------------------------------------------------------
	// public variables
	//--------------------------------------------------------------------
	/* NOTHING */

	//---------------------------------------------------------------------
	// public methods
	//---------------------------------------------------------------------
    static bool Initialize( ID3D11Device* pDevice );
    static bool Release();
    static ID3D11VertexShader* GetDefaultVertexShader();
    static ID3D11PixelShader*  GetDefaultPixelShader();
    static ID3D11InputLayout*  GetDefaultInputLayout();
    static void SetMatrix( 
        ID3D11DeviceContext* pImmediateContext,
        const D3DXMATRIX &world,
        const D3DXMATRIX &view,
        const D3DXMATRIX &proj );
    static void SetVector(
        ID3D11DeviceContext* pImmediateContext,
        const D3DXVECTOR3 &lightDir,
        const D3DXVECTOR3 &viewDir );
};


/////////////////////////////////////////////////////////////////////////////////////
// SubsetData strcture
/////////////////////////////////////////////////////////////////////////////////////
typedef struct _SubsetData
{
    uint32_t			 count;			//!< _JEg
	uint32_t             startIndex;	//!< Jnԍ
    Asura::Math::Vector4 diffuse;		//!< gUːF
    Asura::Math::Vector3 specular;		//!< ʔːF
    Asura::Math::Vector3 emissive;		//!< ȔF
    float				 power;			//!< ʔ˂̋
    ID3D11ShaderResourceView* pSRV;		//!< eNX`f[^
} SubsetData;


/////////////////////////////////////////////////////////////////////////////////////
// MeshData class
/////////////////////////////////////////////////////////////////////////////////////
class MeshData
{
    //-------------------------------------------------------------------------------
    // list of friend classes and methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

private:
    //-------------------------------------------------------------------------------
    // private variables
    //-------------------------------------------------------------------------------
    /* NOTHING */

    //-------------------------------------------------------------------------------
    // private methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

protected:
    //-------------------------------------------------------------------------------
    // protected variables
    //-------------------------------------------------------------------------------
    std::vector< SubsetData >       mSubsets;      //!< TuZbg
    ID3D11Buffer*                   mpVB;          //!< _obt@
    ID3D11Buffer*                   mpMB;          //!< }eAobt@
	ID3D11Buffer*					mpIB;		   //!< CfbNXobt@
    uint32_t                        mOffset;       //!< _obt@̃ItZbg
    uint32_t                        mStride;       //!< _obt@̃XgCh
    
    //-------------------------------------------------------------------------------
    // protected methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

public:
    //-------------------------------------------------------------------------------
    // public variables
    //-------------------------------------------------------------------------------
    /* NOTHING */

    //-------------------------------------------------------------------------------
    // public methods
    //-------------------------------------------------------------------------------
    MeshData();
    ~MeshData();
    void            Release();
    bool            CreateBuffer( ID3D11Device* pDevice, Asura::Graphics::IMesh* pMesh );
    void            Draw( ID3D11DeviceContext* pImmediateContext );
    void            DrawSubset( ID3D11DeviceContext* pImmeidateContext, const int index );
    uint32_t        GetNumSubsets();
    SubsetData&     GetSubset( const int index );
    ID3D11Buffer*   GetVertexBuffer();
    ID3D11Buffer*   GetMaterialBuffer();
	ID3D11Buffer*   GetIndexBuffer();
    uint32_t        GetStride();
    uint32_t        GetOffset();
};


/////////////////////////////////////////////////////////////////////////////////////
// ModelData class
/////////////////////////////////////////////////////////////////////////////////////
class ModelData
{
    //-------------------------------------------------------------------------------
    // list of friend classes and methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

private:
    //-------------------------------------------------------------------------------
    // private variables
    //-------------------------------------------------------------------------------
    /* NOTHING */

    //-------------------------------------------------------------------------------
    // private methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

protected:
    //-------------------------------------------------------------------------------
    // protected variables
    //-------------------------------------------------------------------------------
    std::vector< MeshData >     mMeshes;       //<! bV^Cv
    Asura::Graphics::IModel*    mpInterface;   //!< C^[tF[X

    //-------------------------------------------------------------------------------
    // protected methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

public:
    //-------------------------------------------------------------------------------
    // public variables
    //-------------------------------------------------------------------------------
    /* NOTHING */

    //-------------------------------------------------------------------------------
    // public methods
    //-------------------------------------------------------------------------------
    ModelData();
    virtual ~ModelData();
    virtual bool LoadFromX( const char* filename );
    virtual void Release();
	bool         CreateBuffer( ID3D11Device* pDevice );
	void         ReleaseBuffer();
    void         Draw( ID3D11DeviceContext* pImmediateContext );
    void         DrawMesh( ID3D11DeviceContext* pImmediateConet, const int index );
    uint32_t     GetNumMeshes();
    MeshData&    GetMesh( const int index );
	Asura::Graphics::IModel* GetInterface();
};


//////////////////////////////////////////////////////////////////////////////////////
// SkinningModelData class
//////////////////////////////////////////////////////////////////////////////////////
class SkinningModelData : public ModelData
{
    //-------------------------------------------------------------------------------
    // list of friend classes and methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

private:
    //-------------------------------------------------------------------------------
    // private variables
    //-------------------------------------------------------------------------------
    /* NOTHING */

    //-------------------------------------------------------------------------------
    // private methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

protected:
    //-------------------------------------------------------------------------------
    // protected variables
    //-------------------------------------------------------------------------------
	Asura::Math::Matrix 			   mRoot;
    Asura::Timer					   mTimer;
	Asura::Graphics::IAnimationPlayer* mpPlayer;
	Asura::Graphics::IAnimationClip*   mpCurClip;


    //-------------------------------------------------------------------------------
    // protected methods
    //-------------------------------------------------------------------------------
    /* NOTHING */

public:
    //-------------------------------------------------------------------------------
    // public variables
    //-------------------------------------------------------------------------------
    /* NOTHING */

    //-------------------------------------------------------------------------------
    // public methods
    //-------------------------------------------------------------------------------
    SkinningModelData();
    virtual ~SkinningModelData();
    virtual bool LoadFromX( const char* filename );
    virtual void Release();
    uint32_t     GetNumClips();
    void         SetClip( const uint32_t index );
	void         SetRootMatrix( const Asura::Math::Matrix &matrix );
	void         SetD3DXRootMatrix( const D3DXMATRIX &matrix );
    void         Update( const bool isStop = false, const float speed = 1.0f );

	Asura::Graphics::IAnimationClip*   GetCurrentClip();
	std::vector< Asura::Math::Matrix > GetSkinTransformMatrix();
	std::vector< D3DXMATRIX >          GetD3DXSkinTransformMatrix();
	Asura::Math::Matrix                GetRootMatrix();
	D3DXMATRIX                         GetD3DXRootMatrix();
};


} // namespace Util
} // namespace Asura